@@ -1,8 +1,9 @@
 # pylint:disable=unused-argument,no-self-use
-from typing import Dict, Type, Callable, Any, Optional
+from typing import Any
+from collections.abc import Callable
 
 from . import Block
-from .statement import Call, Statement, ConditionalJump, Assignment, Store, Return
+from .statement import Call, Statement, ConditionalJump, Assignment, Store, Return, Jump
 from .expression import (
     Load,
     Expression,
@@ -16,6 +17,8 @@ from .expression import (
     Register,
     Const,
     MultiStatementExpression,
+    VirtualVariable,
+    Phi,
 )
 
 
@@ -30,6 +33,7 @@ class AILBlockWalkerBase:
             Call: self._handle_Call,
             Store: self._handle_Store,
             ConditionalJump: self._handle_ConditionalJump,
+            Jump: self._handle_Jump,
             Return: self._handle_Return,
         }
 
@@ -46,10 +50,12 @@ class AILBlockWalkerBase:
             Register: self._handle_Register,
             Const: self._handle_Const,
             MultiStatementExpression: self._handle_MultiStatementExpression,
+            VirtualVariable: self._handle_VirtualVariable,
+            Phi: self._handle_Phi,
         }
 
-        self.stmt_handlers: Dict[Type, Callable] = stmt_handlers if stmt_handlers else _default_stmt_handlers
-        self.expr_handlers: Dict[Type, Callable] = expr_handlers if expr_handlers else _default_expr_handlers
+        self.stmt_handlers: dict[type, Callable] = stmt_handlers if stmt_handlers else _default_stmt_handlers
+        self.expr_handlers: dict[type, Callable] = expr_handlers if expr_handlers else _default_expr_handlers
 
     def walk(self, block: Block):
         i = 0
@@ -64,13 +70,13 @@ class AILBlockWalkerBase:
     def walk_expression(
         self,
         expr: Expression,
-        stmt_idx: Optional[int] = None,
-        stmt: Optional[int] = None,
-        block: Optional[Block] = None,
+        stmt_idx: int | None = None,
+        stmt: int | None = None,
+        block: Block | None = None,
     ):
         return self._handle_expr(0, expr, stmt_idx, stmt, block)
 
-    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Optional[Block]) -> Any:
+    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Block | None) -> Any:
         try:
             handler = self.stmt_handlers[type(stmt)]
         except KeyError:
@@ -81,7 +87,7 @@ class AILBlockWalkerBase:
         return None
 
     def _handle_expr(
-        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Optional[Statement], block: Optional[Block]
+        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
     ) -> Any:
         try:
             handler = self.expr_handlers[type(expr)]
@@ -96,75 +102,89 @@ class AILBlockWalkerBase:
     # Default handlers
     #
 
-    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Optional[Block]):
+    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
         self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
         self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
 
-    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Optional[Block]):
+    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Block | None):
+        self._handle_expr(-1, stmt.target, stmt_idx, stmt, block)
         if stmt.args:
             for i, arg in enumerate(stmt.args):
                 self._handle_expr(i, arg, stmt_idx, stmt, block)
 
-    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Optional[Block]):
+    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Block | None):
         self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
         self._handle_expr(1, stmt.data, stmt_idx, stmt, block)
 
-    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Optional[Block]):
+    def _handle_Jump(self, stmt_idx: int, stmt: Jump, block: Block | None):
+        self._handle_expr(0, stmt.target, stmt_idx, stmt, block)
+
+    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None):
         self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
         self._handle_expr(1, stmt.true_target, stmt_idx, stmt, block)
         self._handle_expr(2, stmt.false_target, stmt_idx, stmt, block)
 
-    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Optional[Block]):
+    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Block | None):
         if stmt.ret_exprs:
             for i, ret_expr in enumerate(stmt.ret_exprs):
                 self._handle_expr(i, ret_expr, stmt_idx, stmt, block)
 
-    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Block | None):
         self._handle_expr(0, expr.addr, stmt_idx, stmt, block)
 
-    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Block | None):
+        self._handle_expr(-1, expr.target, stmt_idx, stmt, block)
         if expr.args:
             for i, arg in enumerate(expr.args):
                 self._handle_expr(i, arg, stmt_idx, stmt, block)
 
-    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
         self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
         self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)
 
-    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
         self._handle_expr(0, expr.operand, stmt_idx, stmt, block)
 
-    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement, block: Block | None):
         self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
 
-    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement, block: Block | None):
         self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
         self._handle_expr(1, expr.iftrue, stmt_idx, stmt, block)
         self._handle_expr(2, expr.iffalse, stmt_idx, stmt, block)
 
-    def _handle_Tmp(self, expr_idx: int, expr: Tmp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Tmp(self, expr_idx: int, expr: Tmp, stmt_idx: int, stmt: Statement, block: Block | None):
         pass
 
-    def _handle_Register(self, expr_idx: int, expr: Register, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Register(self, expr_idx: int, expr: Register, stmt_idx: int, stmt: Statement, block: Block | None):
         pass
 
-    def _handle_Const(self, expr_idx: int, expr: Const, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Const(self, expr_idx: int, expr: Const, stmt_idx: int, stmt: Statement, block: Block | None):
         pass
 
+    def _handle_VirtualVariable(
+        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement, block: Block | None
+    ):
+        pass
+
+    def _handle_Phi(self, expr_id: int, expr: Phi, stmt_idx: int, stmt: Statement, block: Block | None):
+        for idx, (_, vvar) in enumerate(expr.src_and_vvars):
+            self._handle_expr(idx, vvar, stmt_idx, stmt, block)
+
     def _handle_MultiStatementExpression(
-        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         for idx, stmt_ in enumerate(expr.stmts):
             self._handle_stmt(idx, stmt_, None)
         self._handle_expr(0, expr.expr, stmt_idx, stmt, block)
 
     def _handle_DirtyExpression(
-        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         self._handle_expr(0, expr.dirty_expr, stmt_idx, stmt, block)
 
     def _handle_VEXCCallExpression(
-        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         for idx, operand in enumerate(expr.operands):
             self._handle_expr(idx, operand, stmt_idx, stmt, block)
@@ -175,12 +195,52 @@ class AILBlockWalker(AILBlockWalkerBase):
     Walks all statements and expressions of an AIL node, and rebuilds expressions, statements, or blocks if needed.
 
     If you need a pure walker without rebuilding, use AILBlockWalkerBase instead.
+
+    :ivar update_block: True if the block should be updated in place, False if a new block should be created and
+                        returned as the result of walk().
+    :ivar replace_phi_stmt: True if you want _handle_Phi be called and vvars potentially replaced; False otherwise.
+                            Default to False because in the most majority cases you do not want vvars in a Phi
+                            variable be replaced.
     """
 
-    def __init__(self, stmt_handlers=None, expr_handlers=None):
+    def __init__(
+        self, stmt_handlers=None, expr_handlers=None, update_block: bool = True, replace_phi_stmt: bool = False
+    ):
         super().__init__(stmt_handlers=stmt_handlers, expr_handlers=expr_handlers)
+        self._update_block = update_block
+        self._replace_phi_stmt = replace_phi_stmt
 
-    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Optional[Block]) -> Any:
+    def walk(self, block: Block) -> Block | None:
+        """
+        Walk the block and rebuild it if necessary. The block will be rebuilt in-place (by updating statements in the
+        original block when self._update_block is set to True), or a new block will be created and returned.
+
+        :param block:   The block to walk.
+        :return:        The new block that is rebuilt, or None if the block is not changed or when self._update_block
+                        is set to True.
+        """
+
+        changed = False
+        new_block: Block | None = None
+
+        i = 0
+        while i < len(block.statements):
+            stmt = block.statements[i]
+            new_stmt = self._handle_stmt(i, stmt, block)
+            if new_stmt is not None:
+                changed = True
+                if not self._update_block:
+                    if new_block is None:
+                        new_block = block.copy(statements=block.statements[:i])
+                    new_block.statements.append(new_stmt)
+            else:
+                if new_block is not None:
+                    new_block.statements.append(stmt)
+            i += 1
+
+        return new_block if changed else None
+
+    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Block | None) -> Any:
         try:
             handler = self.stmt_handlers[type(stmt)]
         except KeyError:
@@ -191,7 +251,7 @@ class AILBlockWalker(AILBlockWalkerBase):
         return None
 
     def _handle_expr(
-        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Optional[Statement], block: Optional[Block]
+        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
     ) -> Any:
         try:
             handler = self.expr_handlers[type(expr)]
@@ -209,7 +269,7 @@ class AILBlockWalker(AILBlockWalkerBase):
     # Default handlers
     #
 
-    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Optional[Block]):
+    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
         changed = False
 
         dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
@@ -227,12 +287,12 @@ class AILBlockWalker(AILBlockWalkerBase):
         if changed:
             # update the statement directly in the block
             new_stmt = Assignment(stmt.idx, dst, src, **stmt.tags)
-            if block is not None:
+            if self._update_block and block is not None:
                 block.statements[stmt_idx] = new_stmt
             return new_stmt
         return None
 
-    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Optional[Block]):
+    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Block | None):
         if stmt.args:
             changed = False
             new_args = []
@@ -262,12 +322,12 @@ class AILBlockWalker(AILBlockWalkerBase):
                     ret_expr=stmt.ret_expr,
                     **stmt.tags,
                 )
-                if block is not None:
+                if self._update_block and block is not None:
                     block.statements[stmt_idx] = new_stmt
                 return new_stmt
         return None
 
-    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Optional[Block]):
+    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Block | None):
         changed = False
 
         addr = self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
@@ -295,12 +355,33 @@ class AILBlockWalker(AILBlockWalkerBase):
                 offset=stmt.offset,
                 **stmt.tags,
             )
-            if block is not None:
+            if self._update_block and block is not None:
                 block.statements[stmt_idx] = new_stmt
             return new_stmt
         return None
 
-    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Optional[Block]):
+    def _handle_Jump(self, stmt_idx: int, stmt: Jump, block: Block | None):
+        changed = False
+
+        target = self._handle_expr(0, stmt.target, stmt_idx, stmt, block)
+        if target is not None and target is not stmt.target:
+            changed = True
+        else:
+            target = stmt.target
+
+        if changed:
+            new_stmt = Jump(
+                stmt.idx,
+                target,
+                target_idx=stmt.target_idx,
+                **stmt.tags,
+            )
+            if self._update_block and block is not None:
+                block.statements[stmt_idx] = new_stmt
+            return new_stmt
+        return None
+
+    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None):
         changed = False
 
         condition = self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
@@ -331,12 +412,12 @@ class AILBlockWalker(AILBlockWalkerBase):
                 false_target_idx=stmt.false_target_idx,
                 **stmt.tags,
             )
-            if block is not None:
+            if self._update_block and block is not None:
                 block.statements[stmt_idx] = new_stmt
             return new_stmt
         return None
 
-    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Optional[Block]):
+    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Block | None):
         if stmt.ret_exprs:
             i = 0
             changed = False
@@ -352,7 +433,7 @@ class AILBlockWalker(AILBlockWalkerBase):
 
             if changed:
                 new_stmt = Return(stmt.idx, new_ret_exprs, **stmt.tags)
-                if block is not None:
+                if self._update_block and block is not None:
                     block.statements[stmt_idx] = new_stmt
                 return new_stmt
         return None
@@ -361,7 +442,7 @@ class AILBlockWalker(AILBlockWalkerBase):
     # Expression handlers
     #
 
-    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Block | None):
         addr = self._handle_expr(0, expr.addr, stmt_idx, stmt, block)
 
         if addr is not None and addr is not expr.addr:
@@ -370,7 +451,7 @@ class AILBlockWalker(AILBlockWalkerBase):
             return new_expr
         return None
 
-    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Block | None):
         changed = False
 
         if expr.args:
@@ -397,7 +478,7 @@ class AILBlockWalker(AILBlockWalkerBase):
 
         return None
 
-    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
         changed = False
 
         operand_0 = self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
@@ -419,7 +500,7 @@ class AILBlockWalker(AILBlockWalkerBase):
             return new_expr
         return None
 
-    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
         new_operand = self._handle_expr(0, expr.operand, stmt_idx, stmt, block)
         if new_operand is not None and new_operand is not expr.operand:
             new_expr = expr.copy()
@@ -427,13 +508,13 @@ class AILBlockWalker(AILBlockWalkerBase):
             return new_expr
         return None
 
-    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement, block: Block | None):
         new_operand = self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
         if new_operand is not None and new_operand is not expr.operand:
             return Convert(expr.idx, expr.from_bits, expr.to_bits, expr.is_signed, new_operand, **expr.tags)
         return None
 
-    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement, block: Optional[Block]):
+    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement, block: Block | None):
         changed = False
 
         cond = self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
@@ -462,8 +543,32 @@ class AILBlockWalker(AILBlockWalkerBase):
             return new_expr
         return None
 
+    def _handle_Phi(self, expr_id: int, expr: Phi, stmt_idx: int, stmt: Statement, block: Block | None) -> Phi | None:
+        if not self._replace_phi_stmt:
+            # fallback to the read-only version
+            return super()._handle_Phi(expr_id, expr, stmt_idx, stmt, block)
+
+        changed = False
+
+        src_and_vvars = None
+        for idx, (src, vvar) in enumerate(expr.src_and_vvars):
+            if vvar is None:
+                if src_and_vvars is not None:
+                    src_and_vvars.append((src, None))
+                continue
+            new_vvar = self._handle_expr(idx, vvar, stmt_idx, stmt, block)
+            if new_vvar is not None and new_vvar is not vvar:
+                changed = True
+                if src_and_vvars is None:
+                    src_and_vvars = expr.src_and_vvars[:idx]
+                src_and_vvars.append((src, new_vvar))
+            elif src_and_vvars is not None:
+                src_and_vvars.append((src, vvar))
+
+        return Phi(expr.idx, expr.bits, src_and_vvars, **expr.tags) if changed else None
+
     def _handle_DirtyExpression(
-        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         new_dirty_expr = self._handle_expr(0, expr.dirty_expr, stmt_idx, stmt, block)
         if new_dirty_expr is not None and new_dirty_expr is not expr.dirty_expr:
@@ -473,7 +578,7 @@ class AILBlockWalker(AILBlockWalkerBase):
         return None
 
     def _handle_VEXCCallExpression(
-        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         changed = False
         new_operands = []
@@ -492,7 +597,7 @@ class AILBlockWalker(AILBlockWalkerBase):
         return None
 
     def _handle_MultiStatementExpression(
-        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Optional[Block]
+        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Block | None
     ):
         changed = False
         new_statements = []
