@@ -1,5 +1,8 @@
 # pylint:disable=arguments-renamed,isinstance-second-argument-not-valid-type,missing-class-docstring
-from typing import Optional, List, TYPE_CHECKING
+from __future__ import annotations
+from enum import IntEnum
+
+from typing import TYPE_CHECKING
 
 try:
     import claripy
@@ -7,7 +10,7 @@ except ImportError:
     claripy = None
 
 from .tagged_object import TaggedObject
-from .utils import get_bits, stable_hash, is_none_or_likeable
+from .utils import get_bits, stable_hash, is_none_or_likeable, is_none_or_matchable
 
 if TYPE_CHECKING:
     from .statement import Statement
@@ -41,6 +44,9 @@ class Expression(TaggedObject):
     def likes(self, atom):  # pylint:disable=unused-argument,no-self-use
         raise NotImplementedError()
 
+    def matches(self, atom):  # pylint:disable=unused-argument,no-self-use
+        return NotImplementedError()
+
     def replace(self, old_expr, new_expr):
         if self is old_expr:
             r = True
@@ -65,7 +71,7 @@ class Atom(Expression):
         "variable_offset",
     )
 
-    def __init__(self, idx, variable, variable_offset=0, **kwargs):
+    def __init__(self, idx, variable=None, variable_offset=0, **kwargs):
         super().__init__(idx, 0, **kwargs)
         self.variable = variable
         self.variable_offset = variable_offset
@@ -97,14 +103,22 @@ class Const(Atom):
         return str(self)
 
     def __str__(self):
-        try:
+        if isinstance(self.value, int):
             return "%#x<%d>" % (self.value, self.bits)
-        except TypeError:
+        elif isinstance(self.value, float):
             return "%f<%d>" % (self.value, self.bits)
+        else:
+            return f"{self.value}<{self.bits}>"
 
     def likes(self, other):
-        return type(self) is type(other) and self.value == other.value and self.bits == other.bits
+        # nan is nan, but nan != nan
+        return (
+            type(self) is type(other)
+            and (self.value is other.value or self.value == other.value)
+            and self.bits == other.bits
+        )
 
+    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -114,7 +128,7 @@ class Const(Atom):
     def sign_bit(self):
         return self.value >> (self.bits - 1)
 
-    def copy(self) -> "Const":
+    def copy(self) -> Const:
         return Const(self.idx, self.variable, self.value, self.bits, **self.tags)
 
 
@@ -143,12 +157,13 @@ class Tmp(Atom):
     def likes(self, other):
         return type(self) is type(other) and self.tmp_idx == other.tmp_idx and self.bits == other.bits
 
+    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
         return stable_hash(("tmp", self.tmp_idx, self.bits))
 
-    def copy(self) -> "Tmp":
+    def copy(self) -> Tmp:
         return Tmp(self.idx, self.variable, self.tmp_idx, self.bits, **self.tags)
 
 
@@ -171,6 +186,8 @@ class Register(Atom):
     def likes(self, atom):
         return type(self) is type(atom) and self.reg_offset == atom.reg_offset and self.bits == atom.bits
 
+    matches = likes
+
     def __repr__(self):
         return str(self)
 
@@ -187,10 +204,217 @@ class Register(Atom):
     def _hash_core(self):
         return stable_hash(("reg", self.reg_offset, self.bits, self.idx))
 
-    def copy(self) -> "Register":
+    def copy(self) -> Register:
         return Register(self.idx, self.variable, self.reg_offset, self.bits, **self.tags)
 
 
+class VirtualVariableCategory(IntEnum):
+    REGISTER = 0
+    STACK = 1
+    MEMORY = 2
+    PARAMETER = 3
+    UNKNOWN = 4
+
+
+class VirtualVariable(Atom):
+
+    __slots__ = (
+        "bits",
+        "varid",
+        "category",
+        "oident",
+    )
+
+    def __init__(
+        self,
+        idx,
+        varid: int,
+        bits,
+        category: VirtualVariableCategory,
+        oident: int | str | tuple | None = None,
+        **kwargs,
+    ):
+        super().__init__(idx, **kwargs)
+
+        self.varid = varid
+        self.category = category
+        self.oident = oident
+        self.bits = bits
+
+    @property
+    def size(self):
+        return self.bits // 8
+
+    @property
+    def was_reg(self) -> bool:
+        return self.category == VirtualVariableCategory.REGISTER
+
+    @property
+    def was_stack(self) -> bool:
+        return self.category == VirtualVariableCategory.STACK
+
+    @property
+    def was_parameter(self) -> bool:
+        return self.category == VirtualVariableCategory.PARAMETER
+
+    @property
+    def reg_offset(self) -> int | None:
+        if self.was_reg:
+            return self.oident
+        return None
+
+    @property
+    def stack_offset(self) -> int | None:
+        if self.was_stack:
+            return self.oident
+        return None
+
+    def likes(self, atom):
+        return (
+            isinstance(atom, VirtualVariable)
+            and self.varid == atom.varid
+            and self.bits == atom.bits
+            and self.category == atom.category
+            and self.oident == atom.oident
+        )
+
+    def matches(self, atom):
+        return (
+            isinstance(atom, VirtualVariable)
+            and self.bits == atom.bits
+            and self.category == atom.category
+            and self.oident == atom.oident
+        )
+
+    def __repr__(self):
+        ori_str = ""
+        match self.category:
+            case VirtualVariableCategory.REGISTER:
+                ori_str = f"{{reg {self.reg_offset}}}"
+            case VirtualVariableCategory.STACK:
+                ori_str = f"{{stack {self.oident}}}"
+        return f"vvar_{self.varid}{ori_str}"
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash(("var", self.varid, self.bits, self.category, self.oident))
+
+    def copy(self) -> VirtualVariable:
+        return VirtualVariable(
+            self.idx,
+            self.varid,
+            self.bits,
+            self.category,
+            oident=self.oident,
+            variable=self.variable,
+            variable_offset=self.variable_offset,
+            **self.tags,
+        )
+
+
+class Phi(Atom):
+
+    __slots__ = (
+        "bits",
+        "src_and_vvars",
+    )
+
+    def __init__(
+        self,
+        idx,
+        bits,
+        src_and_vvars: list[tuple[tuple[int, int], VirtualVariable | None]],
+        **kwargs,
+    ):
+        super().__init__(idx, **kwargs)
+        self.bits = bits
+        self.src_and_vvars = src_and_vvars
+
+    @property
+    def size(self) -> int:
+        return self.bits // 8
+
+    @property
+    def op(self) -> str:
+        return "Phi"
+
+    @property
+    def verbose_op(self) -> str:
+        return "Phi"
+
+    def likes(self, atom) -> bool:
+        if isinstance(atom, Phi) and self.bits == atom.bits:
+            self_src_and_vvarids = {(src, vvar.varid if vvar is not None else None) for src, vvar in self.src_and_vvars}
+            other_src_and_vvarids = {
+                (src, vvar.varid if vvar is not None else None) for src, vvar in atom.src_and_vvars
+            }
+            return self_src_and_vvarids == other_src_and_vvarids
+        return False
+
+    def matches(self, atom) -> bool:
+        if isinstance(atom, Phi) and self.bits == atom.bits:
+            if len(self.src_and_vvars) != len(atom.src_and_vvars):
+                return False
+            self_src_and_vvars = dict(self.src_and_vvars)
+            other_src_and_vvars = dict(atom.src_and_vvars)
+            for src, self_vvar in self_src_and_vvars.items():
+                if src not in other_src_and_vvars:
+                    return False
+                other_vvar = other_src_and_vvars[src]
+                if self_vvar is None and other_vvar is None:
+                    continue
+                if (
+                    self_vvar is None
+                    and other_vvar is not None
+                    or self_vvar is not None
+                    and other_vvar is None
+                    or not self_vvar.matches(other_vvar)
+                ):
+                    return False
+            return True
+        return False
+
+    def __repr__(self):
+        return f"𝜙@{self.bits}b {self.src_and_vvars}"
+
+    __hash__ = TaggedObject.__hash__
+
+    def _hash_core(self):
+        return stable_hash(("phi", self.bits, tuple(sorted(self.src_and_vvars))))
+
+    def copy(self) -> Phi:
+        return Phi(
+            self.idx,
+            self.bits,
+            self.src_and_vvars[::],
+            variable=self.variable,
+            variable_offset=self.variable_offset,
+            **self.tags,
+        )
+
+    def replace(self, old_expr, new_expr):
+        replaced = False
+        new_src_and_vvars = []
+        for src, vvar in self.src_and_vvars:
+            if vvar == old_expr and isinstance(new_expr, VirtualVariable):
+                replaced = True
+                new_src_and_vvars.append((src, new_expr))
+            else:
+                new_src_and_vvars.append((src, vvar))
+
+        if replaced:
+            return True, Phi(
+                self.idx,
+                self.bits,
+                new_src_and_vvars,
+                variable=self.variable,
+                variable_offset=self.variable_offset,
+                **self.tags,
+            )
+        return False, self
+
+
 class Op(Expression):
     __slots__ = ("op",)
 
@@ -211,11 +435,11 @@ class UnaryOp(Op):
         "variable_offset",
     )
 
-    def __init__(self, idx, op, operand, variable=None, variable_offset=None, **kwargs):
+    def __init__(self, idx, op, operand, variable=None, variable_offset=None, bits: int | None = None, **kwargs):
         super().__init__(idx, (operand.depth if isinstance(operand, Expression) else 0) + 1, op, **kwargs)
 
         self.operand = operand
-        self.bits = operand.bits
+        self.bits = operand.bits if bits is None else bits
         self.variable = variable
         self.variable_offset = variable_offset
 
@@ -227,7 +451,18 @@ class UnaryOp(Op):
 
     def likes(self, other):
         return (
-            type(other) is UnaryOp and self.op == other.op and self.bits == other.bits and self.operand == other.operand
+            type(other) is UnaryOp
+            and self.op == other.op
+            and self.bits == other.bits
+            and self.operand.likes(other.operand)
+        )
+
+    def matches(self, atom):
+        return (
+            type(atom) is UnaryOp
+            and self.op == atom.op
+            and self.bits == atom.bits
+            and self.operand.matches(atom.operand)
         )
 
     __hash__ = TaggedObject.__hash__
@@ -236,10 +471,14 @@ class UnaryOp(Op):
         return stable_hash((self.op, self.operand, self.bits))
 
     def replace(self, old_expr, new_expr):
-        r, replaced_operand = self.operand.replace(old_expr, new_expr)
+        if self.operand == old_expr:
+            r = True
+            replaced_operand = new_expr
+        else:
+            r, replaced_operand = self.operand.replace(old_expr, new_expr)
 
         if r:
-            return True, UnaryOp(self.idx, self.op, replaced_operand, **self.tags)
+            return True, UnaryOp(self.idx, self.op, replaced_operand, bits=self.bits, **self.tags)
         else:
             return False, self
 
@@ -251,9 +490,15 @@ class UnaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "UnaryOp":
+    def copy(self) -> UnaryOp:
         return UnaryOp(
-            self.idx, self.op, self.operand, variable=self.variable, variable_offset=self.variable_offset, **self.tags
+            self.idx,
+            self.op,
+            self.operand,
+            variable=self.variable,
+            variable_offset=self.variable_offset,
+            bits=self.bits,
+            **self.tags,
         )
 
     def has_atom(self, atom, identity=True):
@@ -317,6 +562,19 @@ class Convert(UnaryOp):
             and self.rounding_mode == other.rounding_mode
         )
 
+    def matches(self, other):
+        return (
+            type(other) is Convert
+            and self.from_bits == other.from_bits
+            and self.to_bits == other.to_bits
+            and self.bits == other.bits
+            and self.is_signed == other.is_signed
+            and self.operand.matches(other.operand)
+            and self.from_type == other.from_type
+            and self.to_type == other.to_type
+            and self.rounding_mode == other.rounding_mode
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -365,7 +623,7 @@ class Convert(UnaryOp):
         else:
             return False, self
 
-    def copy(self) -> "Convert":
+    def copy(self) -> Convert:
         return Convert(
             self.idx,
             self.from_bits,
@@ -412,7 +670,17 @@ class Reinterpret(UnaryOp):
             and self.from_type == other.from_type
             and self.to_bits == other.to_bits
             and self.to_type == other.to_type
-            and self.operand == other.operand
+            and self.operand.likes(other.operand)
+        )
+
+    def matches(self, other):
+        return (
+            type(other) is Reinterpret
+            and self.from_bits == other.from_bits
+            and self.from_type == other.from_type
+            and self.to_bits == other.to_bits
+            and self.to_type == other.to_type
+            and self.operand.matches(other.operand)
         )
 
     __hash__ = TaggedObject.__hash__
@@ -442,7 +710,7 @@ class Reinterpret(UnaryOp):
         else:
             return False, self
 
-    def copy(self) -> "Reinterpret":
+    def copy(self) -> Reinterpret:
         return Reinterpret(
             self.idx, self.from_bits, self.from_type, self.to_bits, self.to_type, self.operand, **self.tags
         )
@@ -457,15 +725,21 @@ class BinaryOp(Op):
         "variable_offset",
         "floating_point",
         "rounding_mode",
+        "from_bits",  # for divmod
+        "to_bits",  # for divmod
+        "vector_count",
+        "vector_size",
     )
 
     OPSTR_MAP = {
         "Add": "+",
         "AddF": "+",
+        "AddV": "+",
         "Sub": "-",
         "SubF": "-",
         "Mul": "*",
         "MulF": "*",
+        "MulV": "*",
         "Div": "/",
         "DivF": "/",
         "DivMod": "/m",
@@ -490,6 +764,11 @@ class BinaryOp(Op):
         "CmpGTs": ">s",
         "CmpGEs": ">=s",
         "Concat": "CONCAT",
+        "Ror": "ROR",
+        "Rol": "ROL",
+        "Carry": "CARRY",
+        "SCarry": "SCARRY",
+        "SBorrow": "SBORROW",
     }
 
     COMPARISON_NEGATION = {
@@ -514,8 +793,12 @@ class BinaryOp(Op):
         variable=None,
         variable_offset=None,
         bits=None,
-        floating_point=False,
-        rounding_mode=None,
+        floating_point: bool = False,
+        rounding_mode: str | None = None,
+        from_bits: int | None = None,
+        to_bits: int | None = None,
+        vector_count: int | None = None,
+        vector_size: int | None = None,
         **kwargs,
     ):
         depth = (
@@ -540,8 +823,18 @@ class BinaryOp(Op):
             self.bits = bits
         elif self.op == "CmpF":
             self.bits = 32  # floating point comparison
-        elif self.op.startswith("Cmp"):
+        elif self.op in {
+            "CmpEQ",
+            "CmpNE",
+            "CmpLT",
+            "CmpGE",
+            "CmpLE",
+            "CmpGT",
+            "ExpCmpNE",
+        }:
             self.bits = 1
+        elif self.op in {"Carry", "SCarry", "SBorrow"}:
+            self.bits = 8
         elif self.op == "Concat":
             self.bits = get_bits(operands[0]) + get_bits(operands[1])
         elif self.op == "Mull":
@@ -552,7 +845,12 @@ class BinaryOp(Op):
         self.variable = variable
         self.variable_offset = variable_offset
         self.floating_point = floating_point
-        self.rounding_mode = rounding_mode
+        self.rounding_mode: str | None = rounding_mode
+        self.vector_count = vector_count
+        self.vector_size = vector_size
+
+        self.from_bits = from_bits
+        self.to_bits = to_bits
 
         # TODO: sanity check of operands' sizes for some ops
         # assert self.bits == operands[1].bits
@@ -575,6 +873,17 @@ class BinaryOp(Op):
             and self.rounding_mode == other.rounding_mode
         )
 
+    def matches(self, other):
+        return (
+            type(other) is BinaryOp
+            and self.op == other.op
+            and self.bits == other.bits
+            and self.signed == other.signed
+            and is_none_or_matchable(self.operands, other.operands, is_list=True)
+            and self.floating_point == other.floating_point
+            and self.rounding_mode == other.rounding_mode
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -621,24 +930,23 @@ class BinaryOp(Op):
         else:
             r1, replaced_operand_1 = False, None
 
+        r2, replaced_rm = False, None
         if self.rounding_mode is not None:
             if self.rounding_mode == old_expr:
                 r2 = True
                 replaced_rm = new_expr
-            elif isinstance(self.rounding_mode, Expression):
-                r2, replaced_rm = self.rounding_mode.replace(old_expr, new_expr)
-            else:
-                r2, replaced_rm = False, None
 
         if r0 or r1:
             return True, BinaryOp(
                 self.idx,
                 self.op,
-                [replaced_operand_0, replaced_operand_1],
+                [replaced_operand_0 if r0 else self.operands[0], replaced_operand_1 if r1 else self.operands[1]],
                 self.signed,
                 bits=self.bits,
                 floating_point=self.floating_point,
-                rounding_mode=self.rounding_mode,
+                rounding_mode=replaced_rm if r2 else self.rounding_mode,
+                from_bits=self.from_bits,
+                to_bits=self.to_bits,
                 **self.tags,
             )
         else:
@@ -658,7 +966,7 @@ class BinaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "BinaryOp":
+    def copy(self) -> BinaryOp:
         return BinaryOp(
             self.idx,
             self.op,
@@ -669,6 +977,8 @@ class BinaryOp(Op):
             bits=self.bits,
             floating_point=self.floating_point,
             rounding_mode=self.rounding_mode,
+            from_bits=self.from_bits,
+            to_bits=self.to_bits,
             **self.tags,
         )
 
@@ -707,7 +1017,15 @@ class TernaryOp(Op):
             type(other) is TernaryOp
             and self.op == other.op
             and self.bits == other.bits
-            and self.operands == other.operands
+            and is_none_or_likeable(self.operands, other.operands, is_list=True)
+        )
+
+    def matches(self, other):
+        return (
+            type(other) is TernaryOp
+            and self.op == other.op
+            and self.bits == other.bits
+            and is_none_or_matchable(self.operands, other.operands, is_list=True)
         )
 
     __hash__ = TaggedObject.__hash__
@@ -772,7 +1090,7 @@ class TernaryOp(Op):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "TernaryOp":
+    def copy(self) -> TernaryOp:
         return TernaryOp(self.idx, self.op, self.operands[::], bits=self.bits, **self.tags)
 
 
@@ -832,7 +1150,6 @@ class Load(Expression):
     def _likes_addr(self, other_addr):
         if hasattr(self.addr, "likes") and hasattr(other_addr, "likes"):
             return self.addr.likes(other_addr)
-
         return self.addr == other_addr
 
     def likes(self, other):
@@ -845,12 +1162,27 @@ class Load(Expression):
             and self.alt == other.alt
         )
 
+    def _matches_addr(self, other_addr):
+        if hasattr(self.addr, "matches") and hasattr(other_addr, "matches"):
+            return self.addr.matches(other_addr)
+        return self.addr == other_addr
+
+    def matches(self, other):
+        return (
+            type(other) is Load
+            and self._matches_addr(other.addr)
+            and self.size == other.size
+            and self.endness == other.endness
+            and self.guard == other.guard
+            and self.alt == other.alt
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
         return stable_hash(("Load", self.addr, self.size, self.endness))
 
-    def copy(self) -> "Load":
+    def copy(self) -> Load:
         return Load(
             self.idx,
             self.addr,
@@ -901,9 +1233,18 @@ class ITE(Expression):
     def likes(self, atom):
         return (
             type(atom) is ITE
-            and self.cond == atom.cond
-            and self.iffalse == atom.iffalse
-            and self.iftrue == atom.iftrue
+            and self.cond.likes(atom.cond)
+            and self.iffalse.likes(atom.iffalse)
+            and self.iftrue.likes(atom.iftrue)
+            and self.bits == atom.bits
+        )
+
+    def matches(self, atom):
+        return (
+            type(atom) is ITE
+            and self.cond.matches(atom.cond)
+            and self.iffalse.matches(atom.iffalse)
+            and self.iftrue.matches(atom.iftrue)
             and self.bits == atom.bits
         )
 
@@ -952,7 +1293,7 @@ class ITE(Expression):
     def size(self):
         return self.bits // 8
 
-    def copy(self) -> "ITE":
+    def copy(self) -> ITE:
         return ITE(self.idx, self.cond, self.iffalse, self.iftrue, **self.tags)
 
 
@@ -970,6 +1311,7 @@ class DirtyExpression(Expression):
     def likes(self, other):
         return type(other) is DirtyExpression and other.dirty_expr == self.dirty_expr
 
+    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -981,7 +1323,7 @@ class DirtyExpression(Expression):
     def __str__(self):
         return "[D] %s" % str(self.dirty_expr)
 
-    def copy(self) -> "DirtyExpression":
+    def copy(self) -> DirtyExpression:
         return DirtyExpression(self.idx, self.dirty_expr, bits=self.bits, **self.tags)
 
     def replace(self, old_expr, new_expr):
@@ -1025,6 +1367,15 @@ class VEXCCallExpression(Expression):
             and all(op1.likes(op2) for op1, op2 in zip(other.operands, self.operands))
         )
 
+    def matches(self, other):
+        return (
+            type(other) is VEXCCallExpression
+            and other.cee_name == self.cee_name
+            and len(self.operands) == len(other.operands)
+            and self.bits == other.bits
+            and all(op1.matches(op2) for op1, op2 in zip(other.operands, self.operands))
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -1037,7 +1388,7 @@ class VEXCCallExpression(Expression):
         operands_str = ", ".join(repr(op) for op in self.operands)
         return f"{self.cee_name}({operands_str})"
 
-    def copy(self) -> "VEXCCallExpression":
+    def copy(self) -> VEXCCallExpression:
         return VEXCCallExpression(self.idx, self.cee_name, self.operands, bits=self.bits, **self.tags)
 
     def replace(self, old_expr, new_expr):
@@ -1075,7 +1426,7 @@ class MultiStatementExpression(Expression):
         "expr",
     )
 
-    def __init__(self, idx: Optional[int], stmts: List["Statement"], expr: Expression, **kwargs):
+    def __init__(self, idx: int | None, stmts: list[Statement], expr: Expression, **kwargs):
         super().__init__(idx, expr.depth + 1, **kwargs)
         self.stmts = stmts
         self.expr = expr
@@ -1086,7 +1437,20 @@ class MultiStatementExpression(Expression):
         return stable_hash((MultiStatementExpression,) + tuple(self.stmts) + (self.expr,))
 
     def likes(self, other):
-        return type(self) is type(other) and self.stmts == other.stmts and self.expr == other.expr
+        return (
+            type(self) is type(other)
+            and len(self.stmts) == len(other.stmts)
+            and all(s_stmt.likes(o_stmt) for s_stmt, o_stmt in zip(self.stmts, other.stmts))
+            and self.expr.likes(other.expr)
+        )
+
+    def matches(self, atom):
+        return (
+            type(self) is type(atom)
+            and len(self.stmts) == len(atom.stmts)
+            and all(s_stmt.matches(o_stmt) for s_stmt, o_stmt in zip(self.stmts, atom.stmts))
+            and self.expr.matches(atom.expr)
+        )
 
     def __repr__(self):
         return f"MultiStatementExpression({self.stmts}, {self.expr})"
@@ -1127,6 +1491,9 @@ class MultiStatementExpression(Expression):
             )
         return False, self
 
+    def copy(self) -> MultiStatementExpression:
+        return MultiStatementExpression(self.idx, self.stmts[::], self.expr, **self.tags)
+
 
 #
 # Special (Dummy) expressions
@@ -1176,6 +1543,7 @@ class BasePointerOffset(Expression):
             and self.offset == other.offset
         )
 
+    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -1195,7 +1563,7 @@ class BasePointerOffset(Expression):
             return True, BasePointerOffset(self.idx, self.bits, new_base, new_offset, **self.tags)
         return False, self
 
-    def copy(self) -> "BasePointerOffset":
+    def copy(self) -> BasePointerOffset:
         return BasePointerOffset(self.idx, self.bits, self.base, self.offset, **self.tags)
 
 
@@ -1208,7 +1576,7 @@ class StackBaseOffset(BasePointerOffset):
             offset -= 1 << bits
         super().__init__(idx, bits, "stack_base", offset, **kwargs)
 
-    def copy(self) -> "StackBaseOffset":
+    def copy(self) -> StackBaseOffset:
         return StackBaseOffset(self.idx, self.bits, self.offset, **self.tags)
 
 
