@@ -6,7 +6,7 @@ try:
 except ImportError:
     claripy = None
 
-from .utils import stable_hash, is_none_or_likeable
+from .utils import stable_hash, is_none_or_likeable, is_none_or_matchable
 from .tagged_object import TaggedObject
 from .expression import Expression
 
@@ -35,6 +35,12 @@ class Statement(TaggedObject):
             return expr0 is expr1
         return expr0 == expr1
 
+    def likes(self, atom):  # pylint:disable=unused-argument,no-self-use
+        raise NotImplementedError()
+
+    def matches(self, atom):  # pylint:disable=unused-argument,no-self-use
+        return NotImplementedError()
+
 
 class Assignment(Statement):
     """
@@ -58,6 +64,9 @@ class Assignment(Statement):
     def likes(self, other):
         return type(other) is Assignment and self.dst.likes(other.dst) and self.src.likes(other.src)
 
+    def matches(self, other):
+        return type(other) is Assignment and self.dst.matches(other.dst) and self.src.matches(other.src)
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -138,6 +147,16 @@ class Store(Statement):
             and self.endness == other.endness
         )
 
+    def matches(self, other):
+        return (
+            type(other) is Store
+            and self.addr.matches(other.addr)
+            and self.data.matches(other.data)
+            and self.size == other.size
+            and self.guard == other.guard
+            and self.endness == other.endness
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -224,7 +243,7 @@ class Jump(Statement):
         "target_idx",
     )
 
-    def __init__(self, idx, target, target_idx: Optional[int] = None, **kwargs):
+    def __init__(self, idx, target, target_idx: int | None = None, **kwargs):
         super().__init__(idx, **kwargs)
 
         self.target = target
@@ -236,6 +255,9 @@ class Jump(Statement):
     def likes(self, other):
         return type(other) is Jump and is_none_or_likeable(self.target, other.target)
 
+    def matches(self, other):
+        return type(other) is Jump and is_none_or_matchable(self.target, other.target)
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -290,8 +312,8 @@ class ConditionalJump(Statement):
         condition,
         true_target,
         false_target,
-        true_target_idx: Optional[int] = None,
-        false_target_idx: Optional[int] = None,
+        true_target_idx: int | None = None,
+        false_target_idx: int | None = None,
         **kwargs,
     ):
         super().__init__(idx, **kwargs)
@@ -321,6 +343,14 @@ class ConditionalJump(Statement):
             and is_none_or_likeable(self.false_target, other.false_target)
         )
 
+    def matches(self, other):
+        return (
+            type(other) is ConditionalJump
+            and self.condition.matches(other.condition)
+            and is_none_or_matchable(self.true_target, other.true_target)
+            and is_none_or_matchable(self.false_target, other.false_target)
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -408,8 +438,11 @@ class ConditionalJump(Statement):
 
 class Call(Expression, Statement):
     """
-    Call is both an expression and a statement. The return expression of a call is defined as the ret_expr if and only
-    if the callee function has one return expression.
+    Call is both an expression and a statement.
+
+    When used as a statement, it will set ret_expr, fp_ret_expr, or both if both of them should hold return values.
+    When used as an expression, both ret_expr and fp_ret_expr should be None (and should be ignored). The size of the
+    call expression is stored in the bits attribute.
     """
 
     __slots__ = (
@@ -419,6 +452,7 @@ class Call(Expression, Statement):
         "args",
         "ret_expr",
         "fp_ret_expr",
+        "bits",
     )
 
     def __init__(
@@ -430,6 +464,7 @@ class Call(Expression, Statement):
         args=None,
         ret_expr=None,
         fp_ret_expr=None,
+        bits: int | None = None,
         **kwargs,
     ):
         super().__init__(idx, target.depth + 1 if isinstance(target, Expression) else 1, **kwargs)
@@ -440,6 +475,7 @@ class Call(Expression, Statement):
         self.args = args
         self.ret_expr = ret_expr
         self.fp_ret_expr = fp_ret_expr
+        self.bits = bits if bits is not None else ret_expr.bits if ret_expr is not None else None
 
     def likes(self, other):
         return (
@@ -452,6 +488,17 @@ class Call(Expression, Statement):
             and is_none_or_likeable(self.fp_ret_expr, other.fp_ret_expr)
         )
 
+    def matches(self, other):
+        return (
+            type(other) is Call
+            and is_none_or_matchable(self.target, other.target)
+            and self.calling_convention == other.calling_convention
+            and self.prototype == other.prototype
+            and is_none_or_matchable(self.args, other.args, is_list=True)
+            and is_none_or_matchable(self.ret_expr, other.ret_expr)
+            and is_none_or_matchable(self.fp_ret_expr, other.fp_ret_expr)
+        )
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -485,10 +532,6 @@ class Call(Expression, Statement):
 
         return f"Call({self.target}, {s}, ret: {ret_s}, fp_ret: {fp_ret_s})"
 
-    @property
-    def bits(self):
-        return self.ret_expr.bits
-
     @property
     def size(self):
         return self.bits // 8
@@ -522,7 +565,8 @@ class Call(Expression, Statement):
                 r |= r_arg
                 new_args.append(replaced_arg)
 
-        new_ret_expr = None
+        new_ret_expr = self.ret_expr
+        new_bits = self.bits
         if self.ret_expr:
             if self.ret_expr == old_expr:
                 r_ret = True
@@ -531,8 +575,10 @@ class Call(Expression, Statement):
                 r_ret, replaced_ret = self.ret_expr.replace(old_expr, new_expr)
             r |= r_ret
             new_ret_expr = replaced_ret
+            if replaced_ret is not None:
+                new_bits = replaced_ret.bits
 
-        new_fp_ret_expr = None
+        new_fp_ret_expr = self.fp_ret_expr
         if self.fp_ret_expr:
             if self.fp_ret_expr == old_expr:
                 r_ret = True
@@ -551,6 +597,7 @@ class Call(Expression, Statement):
                 args=new_args,
                 ret_expr=new_ret_expr,
                 fp_ret_expr=new_fp_ret_expr,
+                bits=new_bits,
                 **self.tags,
             )
         else:
@@ -565,6 +612,7 @@ class Call(Expression, Statement):
             args=self.args[::] if self.args is not None else None,
             ret_expr=self.ret_expr,
             fp_ret_expr=self.fp_ret_expr,
+            bits=self.bits,
             **self.tags,
         )
 
@@ -586,6 +634,9 @@ class Return(Statement):
     def likes(self, other):
         return type(other) is Return and is_none_or_likeable(self.ret_exprs, other.ret_exprs, is_list=True)
 
+    def matches(self, other):
+        return type(other) is Return and is_none_or_matchable(self.ret_exprs, other.ret_exprs, is_list=True)
+
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -669,7 +720,7 @@ class Label(Statement):
         "block_idx",
     )
 
-    def __init__(self, idx, name: str, ins_addr: int, block_idx: Optional[int] = None, **kwargs):
+    def __init__(self, idx, name: str, ins_addr: int, block_idx: int | None = None, **kwargs):
         super().__init__(idx, **kwargs)
         self.name = name
         self.ins_addr = ins_addr
@@ -678,6 +729,8 @@ class Label(Statement):
     def likes(self, other: "Label"):
         return isinstance(other, Label)
 
+    matches = likes
+
     def _hash_core(self):
         return stable_hash(
             (
